/*--------------------------------------------------------------*/
/* gvtk.c:							*/
/*	Tcl routines for gvtk command-line functions		*/
/* Copyright (c) 2009  Tim Edwards, Open Circuit Design		*/
/*--------------------------------------------------------------*/

#include <stdio.h>
#include <stdarg.h>	/* for va_copy() */
#include <stdlib.h>	/* for atoi() and others */
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <errno.h>

#include <tk.h>

#include "gvtk.h"

Display *dpy;
Colormap cmap;
Globaldata gvobjs;
GVWindowData *areawin;

/*-----------------------*/
/* Tcl 8.4 compatibility */
/*-----------------------*/

#ifndef CONST84
#define CONST84
#endif

/*----------------------------------------------------------------------*/
/* Deal with systems which don't define va_copy().			*/
/*----------------------------------------------------------------------*/

#ifndef HAVE_VA_COPY
  #ifdef HAVE___VA_COPY
    #define va_copy(a, b) __va_copy(a, b)
  #else
    #define va_copy(a, b) a = b
  #endif
#endif

/*----------------------------------------------------------------------*/
/* Reimplement strdup() to use Tcl_Alloc().				*/
/* Note that "strdup" is defined as "Tcl_Strdup" in gvtk.h.		*/
/*----------------------------------------------------------------------*/

char *Tcl_Strdup(const char *s)
{
   char *snew;
   int slen;

   slen = 1 + strlen(s);
   snew = Tcl_Alloc(slen);
   if (snew != NULL)
      memcpy(snew, s, slen);

   return snew;
}

/*--------------------------------------------------------------*/

void gvtk_resizearea(ClientData clientData, XEvent *eventPtr)
{
   GVWindowDataPtr thiswin = (GVWindowDataPtr)clientData;
   int savewidth, saveheight;
   XEvent discard;

   /* DestroyNotify seems to come under StructureNotifyMask. . . */ 
   if (eventPtr->type == DestroyNotify) return;

   savewidth = thiswin->width;
   saveheight = thiswin->height;
   if ((dpy != NULL) && Tk_IsMapped(thiswin->area)) {

      thiswin->width = Tk_Width(thiswin->area);
      thiswin->height = Tk_Height(thiswin->area);
      thiswin->window = Tk_WindowId(thiswin->area);

      if (thiswin->width != savewidth || thiswin->height != saveheight) {
	 if (thiswin->width > thiswin->pixwidth ||
			thiswin->height > thiswin->pixheight) {
	    if (thiswin->width > thiswin->pixwidth)
               thiswin->pixwidth = thiswin->width;
	    if (thiswin->height > thiswin->pixheight)
               thiswin->pixheight = thiswin->height;

            if (thiswin->dbuf != (Pixmap)NULL) Tk_FreePixmap(dpy, thiswin->dbuf);
            thiswin->dbuf = Tk_GetPixmap(dpy, thiswin->window, thiswin->pixwidth,
		thiswin->pixheight, DefaultDepthOfScreen(Tk_Screen(thiswin->area)));

   	    exit_gs(thiswin, True);	// Exit and restart
	    renderpage(thiswin);
	 }
	 else
	    copypage(thiswin);
      }

      /* Flush all resize events from the buffer */

      while (XCheckWindowEvent(dpy, thiswin->window, StructureNotifyMask,
			&discard) == True);
   }
}

/*--------------------------------------------------------------*/

void drawwindow(GVWindowDataPtr thiswin)
{
   XEvent discard;

   if (!Tk_IsMapped(thiswin->area)) return;
   if (thiswin->dbuf != (Pixmap)NULL) copypage(thiswin);
   while (XCheckWindowEvent(dpy, thiswin->window, ExposureMask, &discard) == True);
}

/*--------------------------------------------------------------*/

void gvtk_drawarea(ClientData clientData, XEvent *eventPtr)
{
   GVWindowDataPtr thiswin = (GVWindowDataPtr)clientData;

   Tcl_ServiceAll();
   drawwindow(thiswin);
}


/*--------------------------------------------------------------*/
/* GUI Initialization under Tk					*/
/* First argument is the Tk path name of the drawing window.	*/
/* This function should be called for each new window created.	*/
/*--------------------------------------------------------------*/

int GUI_init(Tcl_Interp *interp, int objc, Tcl_Obj *CONST objv[])
{
   Tk_Window 	tkwind, tktop;
   char 	*gvtopwin;
   int 		locobjc;

   tktop = Tk_MainWindow(interp);
   if (tktop == (Tk_Window)NULL) return TCL_ERROR;

   /* Check if any parameter is a Tk window name */

   locobjc = objc;
   tkwind = (Tk_Window)NULL;
   while (locobjc > 0) {
      gvtopwin = Tcl_GetString(objv[locobjc - 1]);
      tkwind = Tk_NameToWindow(interp, gvtopwin, tktop);
      if (tkwind != (Tk_Window)NULL)
	 break;
      locobjc--;
   }

   if (tkwind != (Tk_Window)NULL)
      gvobjs.topwin = tkwind;
   else
      gvobjs.topwin = tktop;

   dpy = Tk_Display(gvobjs.topwin);
   cmap = Tk_Colormap(gvobjs.topwin);

   return TCL_OK;
}

/*--------------------------------------------------------------*/
/* Create a new window and attach a ghostscript interpreter 	*/
/* to it.							*/
/*--------------------------------------------------------------*/

GVWindowData *new_ghostscript_target(Tcl_Interp *interp, Tk_Window tkwind)
{
   int 		i, locobjc;
   XGCValues	values;   
   char *gvtopwin, *gvdrawwin;
   char winpath[512];
   GVWindowData *newwin, *searchwin;

   /* Check if this window already exists.  If so, return a pointer to it. */

   for (searchwin = gvobjs.windowlist; searchwin != NULL; searchwin =
			searchwin->next)
      if (searchwin->area == tkwind)
	 return searchwin;

   /* Before creating a new window structure, we need to make sure that	*/
   /* the target window is a "simple" window (from tkSimple.c), which 	*/
   /* we do by passing the command "<window_path> cget -commandproc"	*/
   /* and checking if it returns TCL_OK.				*/

   sprintf(winpath, "%s cget -commandproc", Tk_PathName(tkwind));
   if (Tcl_EvalEx(interp, winpath, -1, 0) != TCL_OK) {
      Tcl_SetResult(interp, "Window is the wrong type for rendering!", NULL);
      return NULL;
   }

   /* Create a new window */

   newwin = (GVWindowData *)Tcl_Alloc(sizeof(GVWindowData));

   /* Prepend to linked window list in global data (gvobjs) */
   newwin->next = gvobjs.windowlist;
   gvobjs.windowlist = newwin;
   newwin->area = tkwind;

   /* Set initial values for the display */

   newwin->rotate = 0;
   newwin->zoom = 1.0;
   newwin->offsetx = 0;
   newwin->offsety = 0;
   newwin->fileptr = (FileListPtr)NULL;
   newwin->page = 0;

   /* Make sure the window is mapped */

   Tk_MapWindow(tkwind);
   newwin->window = Tk_WindowId(tkwind);
   newwin->width = Tk_Width(tkwind);
   newwin->height = Tk_Height(tkwind);

   /* Create the pixmap to render into */

   // TO DO:  These should be set from resource file
   newwin->pixwidth = newwin->width;
   newwin->pixheight = newwin->height;

   newwin->dbuf = Tk_GetPixmap(dpy, newwin->window,
		newwin->pixwidth, newwin->pixheight,
                DefaultDepthOfScreen(Tk_Screen(newwin->area)));

   /* Set up event handlers for the drawing area.			*/
   /* There are purposely no callback functions for this window---it 	*/
   /* is defined as type "simple" to keep down the cruft, as I will	*/
   /* define my own event handlers.					*/

   Tk_CreateEventHandler(tkwind, StructureNotifyMask, 
		(Tk_EventProc *)gvtk_resizearea, (ClientData)newwin);
   Tk_CreateEventHandler(tkwind, ExposureMask, 
		(Tk_EventProc *)gvtk_drawarea, (ClientData)newwin);

   /* Create the GC */

   values.foreground = BlackPixel(dpy, DefaultScreen(dpy));
   values.background = WhitePixel(dpy, DefaultScreen(dpy));
   values.graphics_exposures = False;
   newwin->gc = Tk_GetGC(newwin->area, (unsigned long)(GCForeground
		| GCBackground | GCGraphicsExposures), &values);

   /* Start a ghostscript interpreter for this window	*/
   /* (It does not need an input file at this point)	*/
   ghostinit(newwin);

   return newwin;
}

/*----------------------------------------------*/
/* Add a filename to the list.			*/
/* Open the file and scan its document format	*/
/* directives.					*/
/* Return a pointer to the file	structure on	*/
/* success, or NULL on failure.			*/
/*----------------------------------------------*/

FileListPtr new_file(char *filename)
{
   FILE *fps;
   FileListPtr frecord;
   pid_t gsproc;

   /* Check if the file is already loaded */
   /* TO-DO:  Scan file and use inode to check against known files */

   for (frecord = gvobjs.filelist; frecord != NULL; frecord = frecord->next) {
      if (!strcmp(frecord->filename, filename))
	 return frecord;
   }

   /* Attempt to open the file */
   fps = fopen(filename, "r");

   if (fps != NULL) {

      char tmpstr[200], dscnamecmd[200], *tmpfname;

      /* Check the first line of the file to see if it is a PDF */
      /* (Note:  passcan() rewinds fps)				*/
      
      if (fgets(tmpstr, 199, fps) == NULL) {
	 fclose(fps);
	 return NULL;
      }
      if (!strncmp(tmpstr, "%PDF-", 5)) {

	 fclose(fps);

	 tmpfname = tmpnam(NULL);

         sprintf(tmpstr, "-sPDFname=%s", filename);
         sprintf(dscnamecmd, "-sDSCname=%s", tmpfname);

	 if ((gsproc = vfork()) == 0) { 
	    unsetenv("GHOSTVIEW");
            execlp("gs", "gs", "-dNODISPLAY", "-dQUIET", tmpstr, dscnamecmd,
                filename, "pdf2dsc.ps", "-c", "quit", NULL);
	    exit(0);	/* not reached */
	 }
	 waitpid(gsproc, NULL, 0);
	 fps = fopen(tmpfname, "r");
	 if (fgets == NULL) return NULL;

         frecord = (FileListPtr)Tcl_Alloc(sizeof(FileList));
         frecord->filename = Tcl_Strdup(tmpfname);

	 /* Ensure that the temporary file will be removed when */
	 /* the program exits.					*/
	 unlink(tmpfname);
	 frecord->is_pdf = TRUE;
      }
      else {
         frecord = (FileListPtr)Tcl_Alloc(sizeof(FileList));
         frecord->filename = Tcl_Strdup(filename);
	 frecord->is_pdf = FALSE;
      }
      frecord->next = gvobjs.filelist;
      gvobjs.filelist = frecord;

      frecord->fps = fps;

      /* Parse the file and generate the document structure */
      frecord->doc = psscan(fps);
      return frecord;
   }
   return (FileListPtr)NULL;
}

/*--------------------------------------*/
/* Inline the main wrapper prodedure	*/
/*--------------------------------------*/

int gvtk_start(ClientData clientData, Tcl_Interp *interp,
		int objc, Tcl_Obj *CONST objv[])
{
   char *filearg = NULL;
   int result;

   fprintf(stdout, "Starting gvtk under Tcl interpreter\n");

   gvobjs.windowlist = NULL;
   gvobjs.filelist = NULL;
   gvobjs.topwin = (Tk_Window)NULL;
   areawin = (GVWindowData *)NULL;

   /* gvtk initialization routines */

   if (GUI_init(interp, --objc, ++objv) != TCL_OK) {
      Tcl_SetResult(interp, "Invalid or missing top-level windowname"
		" given to start command.\n", NULL);
      return TCL_ERROR;
   }
   globalinit();

   /* All handling of files and windows is taken care of by the wrapper */

   return TCL_OK;
}

/*------------------------------------------------------*/
