/*----------------------------------------------------------------------*/
/* render.c --- Ghostscript rendering of PostScript files		*/
/* Copyright (c) 2002  Tim Edwards, Johns Hopkins University        	*/
/* These routines work only if ghostscript is on the system.		*/
/*----------------------------------------------------------------------*/

#undef GS_DEBUG
// #define GS_DEBUG

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/wait.h>   /* for waitpid() */
#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>

/*------------------------------------------------------------------------*/
/* Local includes                                                         */
/*------------------------------------------------------------------------*/

#include <tk.h>
#include <X11/Xatom.h>

#include "gvtk.h"

/*------------------------------------------------------------------------*/

/*------------------------------------------------------*/
/* Global variables					*/
/*------------------------------------------------------*/

Atom  gv, gvc, gvpage, gvnext, gvdone;

/*--------------------------------------------------------------*/
/* Set the GHOSTVIEW and GHOSTVIEW atoms, and set the GHOSTVIEW	*/
/* environtment	variable.					*/
/*--------------------------------------------------------------*/

void ghostinit_local(GVWindowData *thiswin)
{
   char _STR[256];
#ifdef HAVE_PUTENV
   static char env_str1[128], env_str2[64];
#endif

   sprintf(_STR, "%ld %d %d %d %d %d %g %g %d %d %d %d",
		0L, 0, 0, 0,
		thiswin->pixwidth * 75 / 72,
		thiswin->pixheight * 75 / 72,
		75.0, 75.0, 0, 0, 0, 0);
   XChangeProperty(dpy, thiswin->window, gv, XA_STRING, 8, PropModeReplace,
	_STR, strlen(_STR));
   sprintf(_STR, "%s %d %d", "Color", BlackPixel(dpy, DefaultScreen(dpy)),
		WhitePixel(dpy, DefaultScreen(dpy)));

   XChangeProperty(dpy, thiswin->window, gvc, XA_STRING, 8, PropModeReplace,
	_STR, strlen(_STR));
   XSync(dpy, False);

#ifdef GS_DEBUG
   fprintf(stderr, "setting GHOSTVIEW to %ld %ld\n", (long)thiswin->window,
		(long)thiswin->dbuf);
#endif

#ifdef HAVE_PUTENV
   sprintf(env_str1, "DISPLAY=%s", XDisplayString(dpy));
   putenv(env_str1);
   sprintf(env_str2, "GHOSTVIEW=%ld %ld", (long)thiswin->window, (long)thiswin->dbuf);
   putenv(env_str2);
#else
   setenv("DISPLAY", XDisplayString(dpy), True);
   sprintf(_STR, "%ld %ld", (long)thiswin->window, (long)thiswin->dbuf);
   setenv("GHOSTVIEW", _STR, True);
#endif
}

/*------------------------------------------------------*/
/* Hack, hack, hack.  Structure copied from Tcl source	*/
/*------------------------------------------------------*/

typedef struct PipeState {
    Tcl_Channel channel;     /* Channel associated with this file. */
    void *inFile;            /* Output from pipe. */
    void *outFile;           /* Input to pipe. */
    void *errorFile;         /* Error output from pipe. */
    int numPids;             /* How many processes are attached to this * pipe? */
    Tcl_Pid *pidPtr;         /* The process IDs themselves. Allocated by
                              * the creator of the pipe. */
    int isNonBlocking;       /* Nonzero when the pipe is in nonblocking
                              * mode. Used to decide whether to wait for
                              * the children at close time. */
} PipeState;

/*------------------------------------------------------*/
/* Client message handler.  Called by the Tk		*/
/* client message handler (see below).			*/
/*							*/
/* Returns True if a client message was received from	*/
/* the ghostscript process, False if not.		*/
/*------------------------------------------------------*/

Boolean render_client(GVWindowData *thiswin, XEvent *eventPtr)
{
   if (eventPtr->xclient.message_type == gvpage) {
#ifdef GS_DEBUG
      fprintf(stdout, "Gvtk: Received PAGE message from ghostscript\n");
#endif
      thiswin->gs_mwin = eventPtr->xclient.data.l[0];
      thiswin->gs_state--;
   }
   else if (eventPtr->xclient.message_type == gvdone) {
#ifdef GS_DEBUG
      fprintf(stdout, "Gvtk: Received DONE message from ghostscript\n");
#endif

      thiswin->gs_channel = (Tcl_Channel)NULL;
      thiswin->gs_mwin = 0;
      thiswin->gs_state = GS_INIT;

      /* Attempt to revive the interpreter */
      ghostinit_local(thiswin);
      start_gs(thiswin);
   } 
   else {
      return False;
   }
   return True;
}

/*------------------------------------------------------*/
/* Tk client handler procedure for the above routine.	*/
/*------------------------------------------------------*/

void handle_client(ClientData clientData, XEvent *eventPtr)
{
   GVWindowData *thiswin;
   XAnyEvent *genEventPtr = (XAnyEvent *)eventPtr;

   for (thiswin = gvobjs.windowlist; thiswin != NULL; thiswin = thiswin->next) {
      if (thiswin->window == genEventPtr->window)
	 break;
   }
   if (thiswin == NULL) {
      fprintf(stderr, "Event received from unknown window!\n");
      return;
   }

   if (render_client(thiswin, eventPtr) == False)
      fprintf(stderr, "Gvtk: Received unknown client message type %d\n",
		(int)eventPtr->xclient.message_type);
   else {
      /* Loop if we received an event while rendering the last one */
      if (thiswin->gs_state == GS_PENDING) {
	 thiswin->gs_state--;
	 renderpage(thiswin);
      }
      else
         copypage(thiswin);
   }
}

/*------------------------------------------------------*/
/* Global initialization				*/
/*------------------------------------------------------*/

void globalinit()
{
   gv = XInternAtom(dpy, "GHOSTVIEW", False);
   gvc = XInternAtom(dpy, "GHOSTVIEW_COLORS", False);
   gvpage = XInternAtom(dpy, "PAGE", False);
   gvnext = XInternAtom(dpy, "NEXT", False);
   gvdone = XInternAtom(dpy, "DONE", False);

   Tk_CreateClientMessageHandler((Tk_ClientMessageProc *)handle_client);
}

/*------------------------------------------------------*/
/* Interpreter initialization				*/
/*------------------------------------------------------*/

void ghostinit(GVWindowData *thiswin)
{
   thiswin->gs_state = GS_INIT;
   thiswin->gs_mwin = 0;
   thiswin->gs_channel = (Tcl_Channel)NULL;
   ghostinit_local(thiswin);
   start_gs(thiswin);
}

/*--------------------------------------------------------*/
/* Start a ghostscript process and arrange the I/O pipes  */
/* (Commented lines cause ghostscript to relay its stderr */
/* to gvtk's stderr)				  	  */
/*--------------------------------------------------------*/

void start_gs(GVWindowData *thiswin)
{
   int std_out[2], std_err[2], ret;
   const char *argv[] = {
	"gs", "-sDEVICE=x11alpha", "-dNOPAUSE", "-dQUIET", "-", NULL};
   int argc = 5;

   /* gs is already running */
   if (thiswin->gs_channel != (Tcl_Channel)NULL) {
      fprintf(stderr, "Gvtk: ghostscript is already running!\n");
      return;
   }

   if (thiswin->gs_state != GS_INIT) {
      fprintf(stderr, "Gvtk: ghostscript is not in initial state!\n");
      return;
   }

   thiswin->gs_channel = Tcl_OpenCommandChannel(gvinterp, argc, argv, TCL_STDIN
#ifndef GS_DEBUG
		| TCL_STDERR
#endif
				);	
   if (thiswin->gs_channel == (Tcl_Channel)NULL) {
      fprintf(stderr, "Exec of gs failed\n");
      return;
   }
   Tcl_SetChannelOption(gvinterp, thiswin->gs_channel, "-blocking", "off");

   /* If we are restarting, then we need to reopen the file to display */

   if (thiswin->fileptr != NULL) {
      FileListPtr bgfile;
      bgfile = thiswin->fileptr;
      if (bgfile->filename != NULL && bgfile->fps == NULL) {
	 bgfile->fps = fopen(bgfile->filename, "r");
	 if (bgfile->doc == NULL)
	    psscan(bgfile->fps);
      }
   }
}

/*-------------------------------------------------------*/
/* Copy directly from a file to the ghostscript renderer */
/*-------------------------------------------------------*/

#define BUFLEN 65536

void send_to_gs(GVWindowData *thiswin, FILE *file, long seekpos, long length)
{
   char bytebuf[BUFLEN];
   int i;
   long curpos = seekpos;
   long bytesleft = length;

   while (bytesleft > BUFLEN) {
      fseek(file, curpos, SEEK_SET);
      fread(bytebuf, 1, BUFLEN, file);
      Tcl_Write(thiswin->gs_channel, bytebuf, BUFLEN);
      curpos += BUFLEN;
      bytesleft -= BUFLEN;
   }
   if (bytesleft > 0) {
      fseek(file, curpos, SEEK_SET);
      fread(bytebuf, 1, bytesleft, file);
      Tcl_Write(thiswin->gs_channel, bytebuf, bytesleft);
   }
}

/*------------------------------------------------------*/
/* Send text to the ghostscript renderer.		*/
/* "length" may be negative, in which case it is 	*/
/* computed from strlen() on the buffer.		*/
/*------------------------------------------------------*/

void text_to_gs(GVWindowData *thiswin, char *buffer, int length)
{
   Tcl_Write(thiswin->gs_channel, buffer, length);
}

/*------------------------------------------------------*/
/* Call ghostscript to render the page into the		*/
/* pixmap buffer.					*/
/*------------------------------------------------------*/

int renderpage(GVWindowData *thiswin)
{
   FileListPtr bgfile;
   char _STR[150];
   XEvent event;

   bgfile = thiswin->fileptr;

   if (bgfile == NULL) return -1;
   if (bgfile->doc == NULL) return -1;
   if (bgfile->fps == (FILE *)NULL) return -1;
   if (thiswin->gs_channel == (Tcl_Channel)NULL) return -1;

   /* If an event is already pending, then we set gs_state to inform	*/
   /* the callback procedure that it needs to call renderpage() again	*/
   /* after the current page is finished.  If additional render		*/
   /* come in the interim, we ignore them.				*/

   if (thiswin->gs_state >= GS_PENDING) {
      if (thiswin->gs_state == GS_PENDING)
	 thiswin->gs_state++;
      return 0;
   }

   /* Ask ghostscript to produce the next page. */
   /* If we have not yet received anything, tell gs to run the file */

   if (thiswin->gs_state == GS_INIT) {
      /* Send everything up to page 1, then send the single requested page. */
      send_to_gs(thiswin, bgfile->fps, 0,
			(unsigned int)bgfile->doc->pages[0].begin - 1);
   }

   /* Send next page without rewinding and re-sending header */
   if (bgfile->is_pdf) {
      sprintf(_STR, "\n/pdfshowpage_setpage {gsave %d %d translate %g %g "
		"scale %d rotate} def\n", thiswin->offsetx, thiswin->offsety,
		thiswin->zoom, thiswin->zoom, thiswin->rotate);
   }
   else {
      sprintf(_STR, "\ngsave %d %d translate %g %g scale %d rotate\n",
		thiswin->offsetx, thiswin->offsety,
		thiswin->zoom, thiswin->zoom, thiswin->rotate);
   }
   text_to_gs(thiswin, _STR, -1);
   send_to_gs(thiswin, bgfile->fps, bgfile->doc->pages[thiswin->page].begin,
			bgfile->doc->pages[thiswin->page].len);
   text_to_gs(thiswin, "grestore\n", 9);
   Tcl_Flush(thiswin->gs_channel);

   /* Queue the next "NEXT" message */

   XSync(dpy, False);

   if ((thiswin->gs_state != GS_INIT) && (thiswin->gs_mwin != 0)) {
      thiswin->gs_state = GS_PENDING;
      event.xclient.type = ClientMessage;
      event.xclient.display = dpy;
      event.xclient.window = thiswin->gs_mwin;
      event.xclient.message_type = gvnext;
      event.xclient.format = 32;
      XSendEvent(dpy, thiswin->gs_mwin, False, 0, &event);
      XFlush(dpy);
   }
   else
      thiswin->gs_state = GS_PENDING;

#ifdef GS_DEBUG
   fprintf(stdout, "Rendering next page from file \"%s\" (sent NEXT).\n",
		bgfile->filename);
#endif

   /* The page will be refreshed when we receive confirmation	*/
   /* from ghostscript that the buffer has been rendered.	*/

   return 0;
}

/*------------------------------------------------------*/
/* Copy the rendered page pixmap to the window.		*/
/*------------------------------------------------------*/

int copypage(GVWindowData *thiswin)
{
   /* Don't copy if the buffer is not ready to use */
   if (thiswin->gs_state != GS_READY)
      return -1;

   XCopyArea(dpy, thiswin->dbuf, thiswin->window, thiswin->gc, 0, 
	     thiswin->pixheight - thiswin->height,
	     thiswin->width, thiswin->height, 0, 0);
   return 0;
}

/*------------------------------------------------------*/
/* This routine is a fast pan using XCopyArea instead	*/
/* of having ghostscript re-render the whole page.  It	*/
/* is used to get a real-time update when panning the	*/
/* window.  When the mouse button is released, the	*/
/* page is re-rendered at the new position.		*/
/*------------------------------------------------------*/

int copyoffset(GVWindowData *thiswin, int offx, int offy)
{
   int llx, lly, width, height, destx, desty;

   /* Don't copy in an init state.  All other states are okay. */
   if (thiswin->gs_state == GS_INIT)
      return -1;

   width = thiswin->width;
   height = thiswin->height;
   llx = -offx;
   lly = thiswin->pixheight - height + offy;
   destx = desty = 0;

   if (lly < 0) {
      desty = -lly;
      height += lly;
      lly = 0;
   }
   if (llx < 0) {
      destx = -llx;
      width += llx;
      llx = 0;
   }

   /* Clear areas of the window that don't get copied */
   if (destx > 0)
      XClearArea(dpy, thiswin->window, 0, 0, destx, thiswin->height, False);

   if (desty > 0)
      XClearArea(dpy, thiswin->window, 0, 0, thiswin->width, desty, False);

   if (destx + width < thiswin->width)
      XClearArea(dpy, thiswin->window, destx + width, 0,
		thiswin->width - destx - width, thiswin->height, False);

   if (desty + height < thiswin->height)
      XClearArea(dpy, thiswin->window, 0, desty + height,
		thiswin->width, thiswin->height - desty - height, False);

   XCopyArea(dpy, thiswin->dbuf, thiswin->window, thiswin->gc, llx, 
	     lly, width, height, destx, desty);

   return 0;
}

/*------------------------------------------------------*/
/* Exit ghostscript (and restart, if "restart" is true)	*/
/*------------------------------------------------------*/

int exit_gs(GVWindowData *thiswin, Boolean restart)
{
   FileListPtr bgfile;
   PipeState *pipePtr;
   int pid;

   if (thiswin->gs_channel == (Tcl_Channel)NULL) return -1;	/* gs not running */

   if ((thiswin->gs_state != GS_INIT) && thiswin->fileptr) {

      bgfile = thiswin->fileptr;
      if ((bgfile != NULL) && (bgfile->fps != NULL)) {
         Tcl_Flush(thiswin->gs_channel);
      }
   }

   pipePtr = (PipeState *) Tcl_GetChannelInstanceData(thiswin->gs_channel);
   pid = (unsigned long)pipePtr->pidPtr[0];

   Tcl_Close(gvinterp, thiswin->gs_channel);
#ifdef GS_DEBUG
   fprintf(stderr, "Rudely blowing away process ID %d\n", pid); 
#endif
   kill(pid, SIGTERM);

   thiswin->gs_channel = (Tcl_Channel)NULL;
   thiswin->gs_mwin = 0;
   thiswin->gs_state = GS_INIT;

   if (restart) {
      ghostinit_local(thiswin);
      start_gs(thiswin);
   }

   return 0;
}

/*----------------------------------------------------------------------*/
