#-------------------------------------------------------------------------
# Start of gvtk GUI configuration file.
# This file is sourced by "gvtk.tcl" (".wishrc")
#-------------------------------------------------------------------------
namespace path {::tcl::mathop ::tcl::mathfunc}

# This script sets up all the gvtk windows and callback functions.

#-------------------------------
# Create a new frame
#-------------------------------

proc gvtk::new_frame {name {position 0}} {
  global GVOps

  set drawing ${name}.mainarea.drawing${position}

  # Check if this frame already exists.
  if {![catch {${drawing} page}]} {return {}}

  simple $drawing -bg white -commandproc "config window $drawing"

  # The drawing area.

  grid $drawing -row 0 -column $position -sticky news

  grid rowconfigure ${name}.mainarea 0 -weight 1
  grid columnconfigure ${name}.mainarea $position -weight 1

  #-------------------------------------------------
  # Set default key bindings
  #-------------------------------------------------

  bind ${drawing}  <Return> {catch {page [+ [page] 1]}}
  bind ${drawing}  <Page_Down> {catch {page [+ [page] 1]}}
  bind ${drawing}  <Page_Up> {catch {page [- [page] 1]}}
  bind ${drawing}  <Home> {page 1}
  bind ${drawing}  <End> {page [document pages]}

  #-------------------------------------------------
  # Allow mouse button1-motion to pan the page
  #-------------------------------------------------

  bind ${drawing} <ButtonPress-1> {set GVOps(grab) {%x %y}}
  bind ${drawing} <Button1-Motion> {\
	set dx [- %x [lindex $GVOps(grab) 0]]; \
	set dy [- [lindex $GVOps(grab) 1] %y]; \
	offset $dx $dy copy
	}
  bind ${drawing} <ButtonRelease-1> {\
	set ox [lindex [offset] 0]; \
	set oy [lindex [offset] 1]; \
	set dx [- %x [lindex $GVOps(grab) 0]]; \
	set dy [- [lindex $GVOps(grab) 1] %y]; \
	offset [+ $ox $dx] [+ $oy $dy]\
	}

  #-------------------------------------------------
  # Enter and Focus events switch the current window
  #-------------------------------------------------

  bind ${drawing} <Enter> "${drawing} focus %W"
  bind ${drawing} <FocusIn> "${drawing} focus %W"

  #-------------------------------------------------
  # Allow mouse wheel to zoom in/out
  #-------------------------------------------------

  bind ${drawing} <Button-4> {zoom [* 1.1 [zoom]]}
  bind ${drawing} <Button-5> {zoom [* 0.9091 [zoom]]}

  return $drawing
}


#-------------------------------
# Main gvtk drawing window
#-------------------------------

proc gvtk::new_window { name } {

  global GVTK_VERSION GVTK_REVISION GVOps GVWinOps tcl_platform

  toplevel $name
  wm title $name [string map {.gv GV} $name]
  wm group $name .
  wm protocol $name WM_DELETE_WINDOW \
	"gvtk::closewindow ${name}.mainarea.drawing0"

  # All the internal frames

  frame ${name}.menubar
  frame ${name}.infobar -borderwidth 1 -bg #a07650
  frame ${name}.mainarea -borderwidth 1 -bg #a07650

  grid propagate ${name} false
  grid ${name}.menubar -sticky news -row 0 -column 0
  grid ${name}.mainarea -sticky news -row 1 -columnspan 2
  grid ${name}.infobar -sticky news -row 2 -columnspan 2

  grid rowconfigure ${name} 0 -weight 0
  grid rowconfigure ${name} 1 -weight 1
  grid rowconfigure ${name} 2 -weight 0

  grid columnconfigure ${name} 0 -weight 0
  grid columnconfigure ${name} 1 -weight 1

  set drawing [gvtk::new_frame ${name} 0]

  # The top menu and message bar

  menubutton ${name}.menubar.filebutton -text File \
	-menu ${name}.menubar.filebutton.filemenu \
	-background beige
  menubutton ${name}.menubar.zoombutton -text Zoom \
	-menu ${name}.menubar.zoombutton.zoommenu \
	-background beige
  menubutton ${name}.menubar.pagebutton -text Page \
	-menu ${name}.menubar.pagebutton.pagemenu \
	-background beige

  grid ${name}.menubar.filebutton \
       ${name}.menubar.zoombutton \
       ${name}.menubar.pagebutton \
	-ipadx 10 -sticky news

  # The top message bar

  label ${name}.message -text \
	"Welcome to gvtk v${GVTK_VERSION} rev ${GVTK_REVISION}" \
	-background beige -justify left -anchor w

  grid ${name}.message -row 0 -column 1 -sticky news -ipadx 10

  label ${name}.infobar.message1 -text "Viewing: Page 1" \
	-background beige
  pack ${name}.infobar.message1 -side left -ipadx 6 -fill y

  #-------------------------------------------------
  # Create the menus
  #-------------------------------------------------

  gvtk::makemenus $name

  # Return the name of the drawing window
  return ${drawing}
}

#----------------------------------------------------------------------
# Delete a window.
#----------------------------------------------------------------------

proc gvtk::closewindow {name} {
   set winlist [config windownames]
   if {[llength $winlist] > 1} {
      if {[lsearch $winlist $name] != -1} {
	 config window $name
	 config remove
	 set newwin [lindex [config windownames] 0]
	 config window $newwin
	 destroy [winfo top $name]
      }
   } else {
      quit
   }
}

#----------------------------------------------------------------------
# Create a new window, and set it to the same page as the current one.
#----------------------------------------------------------------------

proc gvtk::forkwindow {} {
   # Find the next unused suffix
   set suffix 0
   foreach win [config windownames] {
      if {[regexp {\.gvtk([0-9]+)} $win match sfix] == 1} {
         if {$sfix > $suffix} {set suffix $sfix}
      }
   }
   incr suffix
   set newname .gvtk${suffix}
   set drawing [gvtk::new_window $newname]
   config window $drawing
}

#----------------------------------------------------------------------

proc gvtk::makemenus {window} {
   global GVOps

   set m [menu ${window}.menubar.filebutton.filemenu -tearoff 0]
   $m add command -label "New Window" -command {gvtk::forkwindow}
   $m add command -label "New Frame" -command \
	[subst {gvtk::new_frame ${window} 1;\
	${window}.mainarea.drawing1 render \[${window}.mainarea.drawing0 render\]}]
   $m add command -label "Close Window" -command \
	"gvtk::closewindow ${window}.mainarea.drawing0"
   $m add separator
   $m add command -label "Load" -command \
	[subst {${window}.mainarea.drawing0 render \
	\[tk_getOpenFile -filetypes {{{PostScript Files} {.ps}} \
		{{Encapsulated PostScript} {.eps}} \
		{{PDF} {.pdf}}} -defaultextension ".ps"\]}]
   $m add separator
   $m add command -label "Quit" -command "quit"

   set m [menu ${window}.menubar.zoombutton.zoommenu -tearoff 0]
   $m add command -label "50%" -command {zoom 0.5}
   $m add command -label "100%" -command {zoom 1.0}
   $m add command -label "150%" -command {zoom 1.5}
   $m add command -label "+50%" -command {zoom [* 1.5 [zoom]]}
   $m add command -label "-33%" -command {zoom [* 0.6667 [zoom]]}

   set m [menu ${window}.menubar.pagebutton.pagemenu -tearoff 0]
   $m add command -label "First Page" -command {page 1 goto}
   $m add command -label "Next Page" -command {catch {page [+ [page] 1] goto}}
   $m add command -label "Previous Page" -command {catch {page [- [page] 1] goto}}
   $m add command -label "Last Page" -command {page [document pages] goto}
}

#----------------------------------------------------------------------
# Tag callback routines
#----------------------------------------------------------------------

proc gvtk::pagecallback {} {
  global GVOps

  if {[info level] <= 1} {
     set dname [document title]
     set plabel [page label]

     # Pick up the top-level window name from $win using regexp.
     set win [config window]
     if {[regexp {(\.[^\t .]+)} $win match topwin] == 1} {
        ${topwin}.message configure -text \
		"Document:  [document title]"
        ${topwin}.infobar.message1 configure -text \
		"Viewing: Page [page] ([page label])"
     } else {
	puts stdout "Cannot get top-level name from $win"
     }
  }
}

#-----------------------------------------------------------------
# Final setup stuff before exiting back to interpreter
#-----------------------------------------------------------------

# This gets rid of the original "wish", in favor of our own window
# (or the gvtkexec, if not called from the tkcon console)

if {[string range [wm title .] 0 3] == "wish"} {
   wm withdraw .
}

if {[string range [wm title .] 0 7] == "gvtkexec"} {
   wm withdraw .
}

#----------------------------------------------------------------------
# Tag callbacks
#----------------------------------------------------------------------

tag page { gvtk::pagecallback }

#-----------------------------------------------------------------
# End of GUI configuration file.  Gvtk continues to load the gvtk
# startup configuration files.
#-----------------------------------------------------------------
